#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul 17 09:11:58 2020

@author: ets
"""

"""
Tools for hydrological forecasting
"""

from ravenpy.models import get_model
import logging
import xarray as xr
import numpy as np
import pandas as pd
import datetime as dt
import warnings

LOGGER = logging.getLogger("PYWPS")


# This function gets model states after running the model (i.e. states at the end of the run).
def get_raven_states(model, **kwds):
    """Get the RAVEN states file (.rvc file) after a model run.

    Parameters
    ----------
    model : {'HMETS', 'GR4JCN', 'MOHYSE', 'HBVEC'}
      Model name.
    kwds : {}
      Model configuration parameters, including the forcing files (ts).

    Returns
    -------
    rvc : {}
      Raven model forcing file

    """
    # Run the model and get the rvc file for future hotstart.
    m = get_model(model)()
    m(overwrite=True, **kwds)
    rvc = m.outputs["solution"]

    return rvc


# Do the actual forecasting step
def perform_forecasting_step(rvc, model, **kwds):
    """
    Function that might be useful eventually to do a forecast from a model setup.
    """
    # kwds includes 'ts', the forecast timeseries data
    # Setup the model
    m = get_model(model)()

    # Force the initial conditions
    m.resume(rvc)

    # Set the parameters, start dates, etc. required to run the model and run
    m(overwrite=True, **kwds)

    return m.q_sim


def perform_climatology_esp(model_name, forecast_date, forecast_duration, **kwds):
    """
    This function takes the model setup and name as well as forecast data and duration and returns
    an ESP forecast netcdf. The data comes from the climatology data and thus there is a mechanism
    to get the correct data from the time series and exclude the current year.

    Parameters
    ----------
    model_name : {'HMETS', 'MOHYSE', 'GR4JCN', 'HBVEC'}
      Model name to instantiate Raven model.
    forecast_date : datetime.datetime
      Date of the forecast issue.
    forecast_duration : int
      Number of days of forecast, forward looking.
    kwds : dict
      Raven model configuration parameters.

    Returns
    -------
    qsims
      Array of streamflow values from the ESP method along with list of member years

    """
    # Get the timeseries
    tsnc = xr.open_dataset(kwds["ts"])

    # Prepare model instance
    m = get_model(model_name)()

    # Now find the periods of time for warm-up and forecast and add to the model keywords as the defaults are failing
    # (nanoseconds datetimes do not like the year 0001...)
    start_date = pd.to_datetime(tsnc["time"][0].values)
    start_date = start_date.to_pydatetime()

    kwds["start_date"] = start_date

    # Forecasting from Feb 29th is not ideal, we will replace with Feb 28th.
    # Should not change much in a climatological forecast.
    if forecast_date.month == 2 and forecast_date.day == 29:
        forecast_date.replace(day=28)

    # Check to make sure forecast date is not in the first year as we need model warm-up.
    # We cannot use timedelta because if the dataset happens to start on a leap
    # year, then the timedelta=365 days will not be robust. (and we cannot use timedelta(years=1)...)
    dateLimit = start_date.replace(year=start_date.year + 1)
    if dateLimit > forecast_date:
        msg = "Forecast date is whithin the warm-up period. Select another forecast date."
        warnings.warn(msg)

    # initialize the array of forecast variables
    qsims = []

    # list of unique years in the dataset:
    avail_years = list(np.unique(tsnc["time.year"].data))

    # Take a copy of the forecast initial date before overwriting in the forecast step.
    forecast_date_main = forecast_date

    # Remove the year that we are forecasting. Or else it's cheating!
    avail_years.remove(forecast_date.year)

    # Update the forecast end-date, which will be the day prior to the forecast date.
    # So forecasts warm-up will be from day 1 in the dataset to the forecast date.
    kwds["end_date"] = forecast_date - dt.timedelta(days=1)

    # Run model to get rvc file after warm-up using base meteo.
    rvc = get_raven_states(model_name, **kwds)

    # We need to check which years are long enough (ex: wrapping years, 365-day forecast starting in
    # September 2015 will need data up to August 2016 at least)
    for years in avail_years:
        if forecast_date.replace(year=years) + dt.timedelta(
            days=forecast_duration - 1
        ) > pd.to_datetime(tsnc["time"][-1].values):
            avail_years.remove(years)
            msg = f"Year {years} has been removed because it is the last year in the dataset and does not cover the " \
                  f"forecast duration."
            warnings.warn(msg)

    # We will iterate this for all forecast years
    for years in avail_years:

        # Replace the forecast period start and end dates with the climatological ESP dates for the
        # current member (year)
        forecast_date = forecast_date.replace(year=years)
        kwds["start_date"] = forecast_date
        kwds["end_date"] = forecast_date + dt.timedelta(days=forecast_duration - 1)

        # Setup the initial states from the warm-up and run the model.
        # Note that info on start/end dates and timeseries are in the kwds.
        m.resume(rvc)
        m(overwrite=True, **kwds)

        # Add member to the ensemble and retag the dates to the real forecast dates
        # (or else we will get dates from the climate dataset that cover all years)
        new_member = m.q_sim.copy(deep=True)
        new_member["time"] = pd.date_range(
            forecast_date_main, periods=forecast_duration
        )
        qsims.append(new_member)

    # Concatenate the members through a new dimension for the members and remove unused dims.
    qsims = xr.concat(qsims, dim="member")
    qsims = qsims.squeeze()

    # Add the number of the forecast year as member ID
    qsims["member"] = (["member"], avail_years)

    return qsims



import datetime as dt
import json
import tempfile

import xarray as xr
# from pywps import Service
# from pywps.tests import assert_response_success
# from raven.processes import RavenHMETSProcess
from ravenpy.models import HMETS

from .common import CFG_FILE, TESTDATA, client_for

# Get path to ncml file for NRCan data.
NRCAN_path = "https://pavics.ouranos.ca/twitcher/ows/proxy/thredds/dodsC/datasets/gridded_obs/nrcan_v2.ncml"

# Temporary path
filepath = tempfile.mkdtemp() + "/NRCAN_ts.nc"

# Get information for given catchment, could be passed in parameter to the function
salmon = xr.open_dataset(TESTDATA["raven-hmets-nc-ts"])
lat = salmon.lat.values[0]
lon = salmon.lon.values[0]

# Start and end dates
start_date = dt.datetime(2006, 1, 1)
end_date = dt.datetime(2007, 12, 31)

# Get data for the covered period including a 1-degree bounding box for good measure. Eventually we will be
# able to take the catchment polygon as a mask and average points residing inside.

ds = (
    xr.open_dataset(NRCAN_path)
    .sel(
        lat=slice(lat + 1, lat - 1),
        lon=slice(lon - 1, lon + 1),
        time=slice(start_date, end_date),
    )
    .mean(dim={"lat", "lon"}, keep_attrs=True)
)
ds.to_netcdf(filepath)


class TestRavenNRCAN:
    def test_simple(self):

        params = (
            9.5019,
            0.2774,
            6.3942,
            0.6884,
            1.2875,
            5.4134,
            2.3641,
            0.0973,
            0.0464,
            0.1998,
            0.0222,
            -1.0919,
            2.6851,
            0.3740,
            1.0000,
            0.4739,
            0.0114,
            0.0243,
            0.0069,
            310.7211,
            916.1947,
        )

        model = HMETS()
        model(
            ts=filepath,
            params=params,
            start_date=start_date,
            end_date=end_date,
            name="Salmon",
            run_name="test-hmets-NRCAN",
            area=4250.6,
            elevation=843.0,
            latitude=54.4848,
            longitude=-123.3659,
            rain_snow_fraction="RAINSNOW_DINGMAN",
            tasmax={"linear_transform": (1.0, -273.15)},
            tasmin={"linear_transform": (1.0, -273.15)},
            pr={"linear_transform": (86400, 0.0)},
        )


class TestRavenNRCANProcess:
    def test_simple(self):
        client = client_for(
            Service(processes=[RavenHMETSProcess(),], cfgfiles=CFG_FILE)
        )

        params = (
            "9.5019, 0.2774, 6.3942, 0.6884, 1.2875, 5.4134, 2.3641, 0.0973, 0.0464, 0.1998, 0.0222, -1.0919, "
            "2.6851, 0.3740, 1.0000, 0.4739, 0.0114, 0.0243, 0.0069, 310.7211, 916.1947"
        )

        datainputs = (
            "ts=files@xlink:href=file://{ts};"
            "params={params};"
            "start_date={start_date};"
            "end_date={end_date};"
            "init={init};"
            "name={name};"
            "run_name={run_name};"
            "area={area};"
            "latitude={latitude};"
            "longitude={longitude};"
            "elevation={elevation};"
            "rain_snow_fraction={rain_snow_fraction};"
            "nc_spec={tasmax};"
            "nc_spec={tasmin};"
            "nc_spec={pr}".format(
                ts=filepath,  # This is a file on disk, need to pass 'g'
                params=params,
                start_date=start_date,
                end_date=end_date,
                init="155,455",
                name="Salmon",
                run_name="test-hmets-NRCAN",
                area="4250.6",
                elevation="843.0",
                latitude=lat,
                longitude=lon,
                rain_snow_fraction="RAINSNOW_DINGMAN",
                tasmax=json.dumps({"tasmax": {"linear_transform": (1.0, -273.15)}}),
                tasmin=json.dumps({"tasmin": {"linear_transform": (1.0, -273.15)}}),
                pr=json.dumps({"pr": {"linear_transform": (86400.0, 0.0)}}),
            )
        )

        resp = client.get(
            service="WPS",
            request="Execute",
            version="1.0.0",
            identifier="raven-hmets",
            datainputs=datainputs,
        )

        assert_response_success(resp)

        # For testing purposes
        # out = get_output(resp.xml)
        # tmp_file, _ = urlretrieve(out['hydrograph'])
        # tmp=xr.open_dataset(tmp_file)
        # plt.plot(tmp['q_sim'])
        # plt.show()



import pytest
import xarray as xr
from xclim.indicators.land._streamflow import fit, stats

from ravenpy.models import Raven

from .common import get_test_data

SALMON_coords = (-123.3659, 54.4848)  # (lon, lat)
RAVEN = Raven


@pytest.fixture
def q_sim_1(tmp_path):
    """A file storing a Raven streamflow simulation over one basin."""

    return get_test_data(
        "hydro_simulations", "raven-gr4j-cemaneige-sim_hmets-0_Hydrographs.nc"
    )[0]


@pytest.fixture
def ts_stats(q_sim_1, tmp_path):
    q = xr.open_dataset(q_sim_1).q_sim
    ts = stats(q, op="max")
    fn = tmp_path / "ts_stats.nc"
    ts.to_netcdf_(fn)
    return fn


@pytest.fixture
def params(ts_stats, tmp_path):
    ds = xr.open_dataset(ts_stats)
    name = list(ds.data_vars.keys()).pop()
    q = ds[name]
    p = fit(q, dist="gumbel_r")
    fn = tmp_path / "fit.nc"
    p.to_netcdf(fn)
    return fn


@pytest.fixture
def era5_hr(tmp_path_factory):
    """Return a netCDF file with hourly ERA5 data at the Salmon location."""
    path = tmp_path_factory.mktemp("ERA5") / "ERA5.nc"

    # Fetch the data and save to disk if the file has not been created yet.
    path.parent.mkdir(exist_ok=True)
    url = "http://pavics.ouranos.ca/twitcher/ows/proxy/thredds/dodsC/datasets/reanalyses/era5.ncml"

    ds = xr.open_dataset(url)
    lon, lat = SALMON_coords
    out = ds.sel(longitude=lon + 360, latitude=lat, method="nearest").sel(
        time=slice("2018-01-01", "2018-12-31")
    )
    out.to_netcdf(path)

    return path



"""Console script for ravenpy."""
import json
import re
import sys

import click

from ravenpy.models.importers import RoutingProductGridWeightImporter


@click.group()
def cli():
    pass


@cli.command()
@click.argument("shp-file-path", type=click.Path(exists=True))
@click.argument("nc-file-path", type=click.Path(exists=True))
@click.option(
    "-d",
    "--dim-names",
    nargs=2,
    type=click.Tuple([str, str]),
    default=RoutingProductGridWeightImporter.DIM_NAMES,
    show_default=True,
    help="Ordered dimension names of longitude (x) and latitude (y) in the NetCDF file.",
)
@click.option(
    "-v",
    "--var-names",
    nargs=2,
    type=click.Tuple([str, str]),
    default=RoutingProductGridWeightImporter.VAR_NAMES,
    show_default=True,
    help="Ordered variable names of longitude (x) and latitude (y) in the NetCDF file.",
)
@click.option(
    "--hru-id-field",
    type=str,
    default=RoutingProductGridWeightImporter.HRU_ID_FIELD,
    show_default=True,
)
@click.option(
    "-g",
    "--gauge-id",
    multiple=True,
    type=str,
    show_default=True,
    help="Basins of interest (corresponds to 'Gauge-ID' in the shapefile).",
)
@click.option(
    "-s",
    "--sub-id",
    multiple=True,
    type=str,  # to support alternate "-s 123,456" legacy syntax
    show_default=True,
    help="Sub IDs of most downstream subbasins (containing usually a gauge station, corresponds to 'SubId' in the shapefile).",
)
@click.option(
    "-o",
    "--output",
    type=click.Choice(["raven", "json", "text"]),
    default="raven",
    show_default=True,
)
def generate_grid_weights(
    shp_file_path,
    nc_file_path,
    dim_names,
    var_names,
    hru_id_field,
    gauge_id,
    sub_id,
    output,
):
    """
    Generate grid weights in various formats.

    SHP_FILE_PATH: Shapefile that contains all information of the routing toolbox for the catchment of interest (and maybe some more catchments).

    NC_FILE_PATH: NetCDF file containing at least 2D latitudes and 2D longitudes. Grid needs to be representative of model outputs that are then required to be routed.
    """

    # Although Click does not support it, we want to support legacy syntax (e.g. "-s 123,456") for both gauge and sub IDs.
    # The Click way for `multiple` options would be: "-s 123 -s 456"

    if len(gauge_id) == 1:
        gauge_ids = list(filter(None, re.split("\W+", gauge_id[0])))
    else:
        gauge_ids = gauge_id

    if len(sub_id) == 1:
        sub_ids = list(map(int, filter(None, re.split("\W+", sub_id[0]))))
    else:
        sub_ids = list(map(int, sub_id))

    importer = RoutingProductGridWeightImporter(
        shp_file_path,
        nc_file_path,
        dim_names,
        var_names,
        hru_id_field,
        gauge_ids,
        sub_ids,
    )
    cmd = importer.extract()

    if output == "raven":
        print(cmd.to_rv())
    elif output == "text":
        print("\n".join(" ".join(map(str, d)) for d in cmd.data))
    elif output == "json":
        print(
            json.dumps(
                {
                    "NumberHRUs": cmd.number_hrus,
                    "NumberGridCells": cmd.number_grid_cells,
                    "Weights": cmd.data,
                },
                indent=4,
            )
        )


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



import datetime as dt

import pytest

from ravenpy.models import HMETS


@pytest.mark.slow
class TestRavenERA5:
    def test_simple(self, era5_hr):

        params = (
            9.5019,
            0.2774,
            6.3942,
            0.6884,
            1.2875,
            5.4134,
            2.3641,
            0.0973,
            0.0464,
            0.1998,
            0.0222,
            -1.0919,
            2.6851,
            0.3740,
            1.0000,
            0.4739,
            0.0114,
            0.0243,
            0.0069,
            310.7211,
            916.1947,
        )

        model = HMETS()
        model(
            ts=era5_hr,
            params=params,
            start_date=dt.datetime(2018, 1, 1),
            end_date=dt.datetime(2018, 8, 10),
            name="Salmon",
            run_name="test-hmets-era5",
            area=4250.6,
            elevation=843.0,
            latitude=54.4848,
            longitude=-123.3659,
            rain_snow_fraction="RAINSNOW_DINGMAN",
            tas={"linear_transform": (1.0, -273.15), "time_shift": -0.25},
            pr={"linear_transform": (24000.0, 0.0), "time_shift": -0.25},
        )



"""Top-level package for RavenPy."""

import os
import sys
import warnings
from pathlib import Path

from .__version__ import __author__, __email__, __version__  # noqa: F401

# Note: all the binaries associated with this package are installed in the "bin"
#       folder of the target venv, which corresponds to either `sys.prefix` or
#       $CONDA_PREFIX.


if os.getenv("CONDA_PREFIX"):
    VENV_PATH = Path(os.getenv("CONDA_PREFIX"))
elif sys.base_prefix != sys.prefix:
    VENV_PATH = Path(sys.prefix)
else:
    raise IOError("Please run RavenPy in a virtual environment!")


RAVEN_EXEC_PATH = VENV_PATH / "bin" / "raven"
OSTRICH_EXEC_PATH = VENV_PATH / "bin" / "ostrich"


# if "DO_NOT_CHECK_EXECUTABLE_EXISTENCE" not in os.environ:
#     raven_exec = VENV_PATH / "bin" / "raven"
#     if not raven_exec.exists():
#         raise IOError(
#             "The raven executable is not installed in the virtual environment."
#         )

#     ostrich_exec = VENV_PATH / "bin" / "ostrich"
#     if not ostrich_exec.exists():
#         raise IOError(
#             "The ostrich executable is not installed in the virtual environment."
#         )

# raven_simg = VENV_PATH / "bin" / "hydro-raven-latest.simg"
# if not raven_simg.exists():
#     warnings.warn(
#         "The Raven Singularity image has not been downloaded. Execute \n"
#         "$ singularity pull shub://132.217.141.54/hydro/raven:latest \n"
#         "and store the image in <your_venv>/bin/"
#     )



# -*- coding: utf-8 -*-

# This information is located in its own file so that it can be loaded
# without importing the main package when its dependencies are not installed.
# See: https://packaging.python.org/guides/single-sourcing-package-version

__author__ = """David Huard"""
__email__ = 'huard.david@ouranos.ca'
__version__ = '0.10.0'



from click.testing import CliRunner

from ravenpy.cli import generate_grid_weights

from .common import get_test_data


def test_generate_grid_weights():
    runner = CliRunner()
    ds = map(
        str,
        [
            get_test_data("raven-routing-sample", "finalcat_hru_info.zip")[0],
            get_test_data("raven-routing-sample", "VIC_streaminputs.nc")[0],
        ],
    )

    result = runner.invoke(generate_grid_weights, ds)

    assert result.exit_code == 0
    assert not result.exception
    assert ":NumberHRUs 51" in result.output
    assert ":NumberGridCells 100" in result.output
    assert len(result.output.split("\n")) == 216



#!/usr/bin/env python

"""The setup script."""

import os
import shutil
import subprocess
import sys
import urllib.request
import zipfile
from pathlib import Path

from setuptools import find_packages, setup
from setuptools.command.install import install

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "Click~=7.0",
    "gdal==3.0.4",
    "rasterio",
    "statsmodels",
    "xarray==0.16",
    "xclim==0.18",
    "wheel",
]

# setup_requirements = [
#     "pytest-runner",
# ]

# test_requirements = [
#     "pytest>=3",
# ]

# docs_requirements = [
#     dependency for dependency in open("requirements_docs.txt").readlines()
# ]

# dev_requirements = [
#     dependency for dependency in open("requirements_dev.txt").readlines()
# ]


class InstallExternalDeps(install):
    """
    Custom handler for the 'install' command, to download, extract and compile
    the source code of Raven and OSTRICH and copy the resulting binaries in the
    "bin" folder of the current venv.
    """

    user_options = install.user_options + [
        # The format is (long option, short option, description).
        ("with-raven", None, "Download Raven and OSTRICH sources and compile them."),
        ("with-testdata", None, "Download RavenPy test data used for unit tests."),
    ]

    def initialize_options(self):
        """Set default values for options."""
        # Each user option must be listed here with their default value.
        install.initialize_options(self)
        self.with_raven = False
        self.with_testdata = False

    def finalize_options(self):
        install.finalize_options(self)

    def install_binary_dep(
        self, url, name, rev_name, binary_name, venv_path, make_target=""
    ):
        print(f"Downloading {name} source code..")
        urllib.request.urlretrieve(
            f"{url}/{rev_name}.zip", self.external_deps_path / f"{name}.zip"
        )

        print(f"Extracting {name} source code..")
        with zipfile.ZipFile(self.external_deps_path / f"{name}.zip", "r") as zip_ref:
            zip_ref.extractall(self.external_deps_path)

        print(f"Compiling {name}..")
        try:
            subprocess.check_call(
                f"make {make_target}",
                cwd=self.external_deps_path / rev_name,
                shell=True,
            )
        except subprocess.CalledProcessError as e:
            print(e)
            exit(f"There was an error while compiling {name}")

        #  Copy binary into venv bin folder (so it should be in the path when the venv is active)
        target_bin_path = venv_path / "bin" / name
        print(f"Copying binary to {target_bin_path}")
        shutil.copy(
            self.external_deps_path / rev_name / binary_name,
            target_bin_path,
        )

    def run(self):
        venv_path = None
        if os.getenv("CONDA_PREFIX"):
            # Conda env
            venv_path = Path(os.getenv("CONDA_PREFIX"))
        elif sys.base_prefix != sys.prefix:
            # Regular venv
            venv_path = Path(sys.prefix)
        else:
            exit("Error: Please install RavenPy in a virtual environment!")

        if self.with_raven:
            self.external_deps_path = Path("./external_deps")
            self.external_deps_path.mkdir(exist_ok=True)

            url = "http://www.civil.uwaterloo.ca/jmai/raven/"
            self.install_binary_dep(
                url, "raven", "Raven-rev288", "Raven.exe", venv_path
            )
            self.install_binary_dep(
                url,
                "ostrich",
                "Ostrich_2017-12-19_plus_progressJSON",
                f"OstrichGCC",
                venv_path,
                "GCC",
            )

        if self.with_testdata:
            local_zip_path = venv_path / "raven-testdata-master.zip"

            print(f"Downloading raven-tesdata..")
            urllib.request.urlretrieve(
                "https://github.com/Ouranosinc/raven-testdata/archive/master.zip",
                local_zip_path,
            )

            print(f"Extracting raven-testdata to {venv_path}..")
            with zipfile.ZipFile(local_zip_path) as zip_ref:
                zip_ref.extractall(venv_path)

        # This works with python setup.py install, but produces this error with pip install:
        # ERROR: ravenpy==0.1.0 did not indicate that it installed an .egg-info directory. Only setup.py projects generating .egg-info directories are supported.
        # super().do_egg_install()

        # This works with pip install, but has the problem that it ignores install_requires
        # when running with `python setup.py install`:
        # https://stackoverflow.com/questions/21915469/python-setuptools-install-requires-is-ignored-when-overriding-cmdclass
        install.run(self)


setup(
    author="David Huard",
    author_email="huard.david@ouranos.ca",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="A Python wrapper to setup and run the hydrologic modelling framework Raven.",
    entry_points={
        "console_scripts": [
            "ravenpy=ravenpy.cli:cli",
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/x-rst",
    include_package_data=True,
    keywords="ravenpy",
    name="ravenpy",
    packages=find_packages(
        include=[
            "ravenpy",
            "ravenpy.*",
        ],
    ),
    # setup_requires=setup_requirements,
    test_suite="tests",
    # tests_require=test_requirements,
    # extras_require={
    #     "docs": docs_requirements,
    #     "dev": dev_requirements,
    # },
    url="https://github.com/CSHS-CWRA/ravenpy",
    version="0.1.1",
    zip_safe=False,
    cmdclass={"install": InstallExternalDeps},
)



"""
Library to perform graphs for the streamflow time series analysis.

The following graphs can be plotted:
    - hydrograph
    - mean_annual_hydrograph
    - spaghetti_annual_hydrograph

"""

import numpy as np
import pandas as pd
import xarray as xr
from matplotlib import pyplot as plt
from scipy import stats

from ravenpy.utilities.mk_test import mk_test_calc
from xclim.core.units import units2pint


def hydrograph(file_list):
    """
    annual_hydrograph

    INPUTS:
        file_list -- Raven output files containing simulated streamflows

    Create a graphic of the hydrograph for each model simulation.
    """

    ds = [xr.open_dataset(file) for file in file_list]

    # Get time data for the plot
    dates = pd.DatetimeIndex(ds[0].time.values)
    first_date = dates.min().strftime('%Y/%m/%d')
    last_date = dates.max().strftime('%Y/%m/%d')

    basin_name = ds[0].basin_name.values[0]  # selected basin name

    fig, ax = plt.subplots()  # initialize figure

    # Plot the observed streamflows if available
    if hasattr(ds[0], 'q_obs'):
        q_obs = ds[0].q_obs
        plt.plot(
            dates,
            q_obs,
            linewidth=2,
            label='obs')

    # Plot the simulated streamflows for each hydrological model
    q_sim = [h.q_sim for h in ds]
    for sim in q_sim:
        plt.plot(
            dates,
            sim,
            linewidth=2,
            label='sim: ' + basin_name)

    # plt.xlim([first_date, last_date])
    plt.ylim(bottom=0, top=None)
    ax.set_xlabel('Time')
    ax.set_ylabel(r'$Streamflow [m^3s^{{-1}}]$')
    ax.set_title('Hydrograph between {} and {}\nSelected basin: {} basin.'.format(first_date, last_date, basin_name))
    ax.legend()
    ax.grid()

    plt.xticks(rotation=90)
    plt.tight_layout()

    return fig


def mean_annual_hydrograph(file_list):
    """
    mean_annual_hydrograph

    INPUTS:
        file_list -- Raven output files containing simulated streamflows

    Create a graphic of the mean hydrological cycle for each model simulation.
    """

    # Time series for the plot
    ds = [xr.open_dataset(file) for file in file_list]

    # Get time data for the plot
    dates = pd.DatetimeIndex(ds[0].time.values)
    first_date = dates.min().strftime('%Y/%m/%d')
    last_date = dates.max().strftime('%Y/%m/%d')

    basin_name = ds[0].basin_name.values[0]  # selected basin name

    fig, ax = plt.subplots()  # initialize figure

    # Plot the observed streamflows if available
    if hasattr(ds[0], 'q_obs'):
        q_obs = ds[0].q_obs
        mah_obs = q_obs.groupby('time.dayofyear').mean()

        plt.plot(
            mah_obs.dayofyear,
            mah_obs,
            linewidth=2,
            label='obs')

    # Plot the simulated streamflows for each hydrological model
    q_sim = [h.q_sim for h in ds]
    mahs_sim = [q.groupby('time.dayofyear').mean() for q in q_sim]

    for mah in mahs_sim:
        plt.plot(
            mah.dayofyear,
            mah,
            linewidth=2,
            label='sim: ' + basin_name)

    plt.xticks(
        np.linspace(0, 365, 13)[:-1], (
            'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
            'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec',
        ),
    )

    plt.xlim(0, mah_obs.shape[0])
    plt.ylim(bottom=0, top=None)
    ax.set_xlabel('Time')
    ax.set_ylabel(r'$Streamflow [m^3s^{{-1}}]$')
    ax.set_title('Hydrograph between {} and {}\nSelected basin: {} basin.'.format(first_date, last_date, basin_name))
    ax.legend()
    ax.grid()

    plt.tight_layout()

    return fig


def spaghetti_annual_hydrograph(file):
    """
    spaghetti_annual_hydrograph

    INPUTS:
        file -- Raven output file containing simulated streamflows of one model

    Create a spaghetti plot of the mean hydrological cycle for one model
    simulations. The mean simulation is also displayed.
    """

    # Time series for the plot
    ds = xr.open_dataset(file)

    # Get time data for the plot
    dates = pd.DatetimeIndex(ds.time.values)
    first_date = dates.min().strftime('%Y/%m/%d')
    last_date = dates.max().strftime('%Y/%m/%d')

    basin_name = ds.basin_name.values[0]  # selected basin name

    fig, ax = plt.subplots()  # initialize figure

    # Plot the observed streamflows if available
    if hasattr(ds, 'q_obs'):
        q_obs = ds.q_obs
        mah_obs = q_obs.groupby('time.year')
        mah_obs_mean = q_obs.groupby('time.dayofyear').mean()

        for year in mah_obs.groups.keys():
            plt.plot(
                np.arange(1, q_obs.values[mah_obs.groups[year]].shape[0] + 1, 1),
                q_obs.values[mah_obs.groups[year]],
                linewidth=1,
                color='C0')

        plt.plot(
            mah_obs_mean.dayofyear,
            mah_obs_mean,
            linewidth=2,
            color='C0',
            label='obs')

    # Plot the simulated streamflows for each hydrological model
    q_sim = ds.q_sim
    mah_sim = q_sim.groupby('time.year')
    mah_sim_mean = q_sim.groupby('time.dayofyear').mean()

    for year in mah_sim.groups.keys():
        plt.plot(
            np.arange(1, q_sim.values[mah_sim.groups[year]].shape[0] + 1, 1),
            q_sim.values[mah_sim.groups[year]],
            linewidth=1,
            color='C1')

    plt.plot(
        mah_sim_mean.dayofyear,
        mah_sim_mean,
        linewidth=2,
        color='C1',
        label='sim: ' + '<model_name>')

    plt.xticks(
        np.linspace(0, 365, 13)[:-1], (
            'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
            'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec',
        ),
    )

    plt.xlim(0, mah_sim_mean.shape[0])
    plt.ylim(bottom=0, top=None)
    ax.set_xlabel('Time')
    ax.set_ylabel(r'$Streamflow [m^3s^{{-1}}]$')
    ax.set_title('Spaghetti annual hydrograph between {} and {}'
                 '\n Selected basin: {} basin.'.format(first_date, last_date, basin_name))
    ax.legend()
    ax.grid()

    plt.tight_layout()

    return fig


def ts_graphs(file, trend=True, alpha=0.05):
    """
    graphs for time series statistics

    INPUTS:
        file -- xclim file containing streamflow statistics for one run

    Create a figure with the statistics so one can see a trend in the data
    """

    # Time series for the plot
    ds = xr.open_dataset(file)
    titlename = ds['ts_stats'].description
    values = ds['ts_stats'].values[:]

    values[0] = 2
    values[1] = 1
    values[2] = 3
    values[3] = 3
    values[4] = 3
    values[5] = 5
    values[6] = 6
    values[7] = 7
    values[8] = 10
    values[9] = 4

    # Get time data for the plot
    dates = pd.DatetimeIndex(ds.time.values)
    # first_date = dates.min()#.strftime('%Y/%m/%d')
    # last_date = dates.max()#.strftime('%Y/%m/%d')
    res = None

    if trend:
        res = stats.theilslopes(values, dates)
        trd, h, p, z = mk_test_calc(values, alpha=alpha)
        titlename = titlename + ", Mann-Kendall h=" + str(h) + ", p-value=" + str(np.round(p, 4))

    fig, ax = plt.subplots()
    ax.plot(dates, values, label='time-series index')

    # plt.xlim([first_date, last_date])
    ax.set_xlabel('Time')
    ax.set_ylabel(r'$Streamflow [m^3s^{{-1}}]$')

    ax.set_title(titlename)

    ax.grid()
    plt.tight_layout()

    if trend:
        # TODO: This does not work yet, trying to compute the y-value of a datetime x-axis value * a slope...
        ax.plot(dates, res[1] + res[0] * dates,
                linestyle='--',
                linewidth=2,
                label='Sen\'s Slope')

    ax.legend()

    return fig


def ts_fit_graph(ts, params):
    """Create graphic showing an histogram of the data and the distribution fitted to it.

    Parameters
    ----------
    ts : str
      Path to netCDF file storing the time series.
    params : str
      Path to netCDF file storing the distribution parameters.

    Returns
    -------
    fig
    """
    from xclim.indices.generic import get_dist

    n = ts.nbasins.size
    dist = params.attrs['scipy_dist']

    fig, axes = plt.subplots(n, figsize=(10, 6), squeeze=False)

    for i in range(n):
        ax = axes.flat[i]
        ax2 = plt.twinx(ax)
        p = params.isel(nbasins=i)

        # Plot histogram of time series as density then as a normal count.
        density, bins, patches = ax.hist(ts.isel(nbasins=i).dropna(dim='time'),
                                         alpha=.5, density=True, bins='auto',
                                         label="__nolabel__")
        ax2.hist(ts.isel(nbasins=i).dropna(dim='time'), bins=bins, facecolor=(1, 1, 1, 0.01), edgecolor='gray',
                 linewidth=1, )

        # Plot pdf of distribution
        dc = get_dist(dist)(*params.isel(nbasins=i))
        mn = dc.ppf(.01)
        mx = dc.ppf(.99)
        q = np.linspace(mn, mx, 200)
        pdf = dc.pdf(q)
        ps = ', '.join(["{:.1f}".format(x) for x in p.values])
        ax.plot(q, pdf, '-', label="{}({})".format(params.attrs['scipy_dist'], ps))

        # Labels
        ax.set_xlabel("{} (${:~P}$)".format(ts.long_name, units2pint(ts.units)))
        ax.set_ylabel("Probability density")
        ax2.set_ylabel("Histogram count")

        ax.legend(frameon=False)

    plt.tight_layout()
    return fig
